#ifndef TYPON_FUNDAMENTAL_RINGBUFFER_HPP_INCLUDED
#define TYPON_FUNDAMENTAL_RINGBUFFER_HPP_INCLUDED

#include <atomic>


namespace typon::fdt::lock_free
{

  template <typename T>
    requires std::is_trivially_copyable_v<T>
  struct ring_buffer
  {
    using u8 = std::uint_least8_t;
    using u64 = std::uint_fast64_t;

    using enum std::memory_order;

    const u8 _bits;
    const u64 _mask;
    ring_buffer * const _next;
    std::atomic<T> * const _array;

    ring_buffer(u8 bits, ring_buffer * next = nullptr) noexcept
      : _bits(bits)
      , _mask(this->capacity() - 1)
      , _next(next)
      , _array(new std::atomic<T>[this->capacity()])
    {}

    ~ring_buffer()
    {
      delete [] _array;
      if (_next)
      {
        delete _next;
      }
    }

    u64 capacity() noexcept
    {
      return u64(1) << _bits;
    }

    void put(u64 index, T object) noexcept
    {
      _array[index & _mask].store(std::move(object), relaxed);
    }

    T get(u64 index) noexcept
    {
      return _array[index & _mask].load(relaxed);
    }

    ring_buffer * fill(ring_buffer * sink, u64 start, u64 end) noexcept
    {
      for (u64 i = start; i < end; i++)
      {
        sink->put(i, get(i));
      }
      return sink;
    }

    ring_buffer * grow(u64 start, u64 end) noexcept
    {
      return fill(new ring_buffer(_bits + 1, this), start, end);
    }

    ring_buffer * shrink(u64 start, u64 end) noexcept
    {
      return fill(std::exchange(_next, nullptr), start, end);
    }
  };

}


#endif //  TYPON_FUNDAMENTAL_RINGBUFFER_HPP_INCLUDED
